import json
import logging
import platform
import shutil
import subprocess
import sys
import os
import re

import requests
from tqdm import tqdm
from bs4 import BeautifulSoup

from aniworld.config import (
    DEFAULT_REQUEST_TIMEOUT,
    MPV_DIRECTORY,
    ANIWORLD_TO,
    MPV_SCRIPTS_DIRECTORY
)


def check_avx2_support() -> bool:
    if platform.system() != "Windows":
        logging.debug("AVX2 check is only supported on Windows.")
        return False

    import cpuinfo  # type: ignore # pylint: disable=import-outside-toplevel

    info = cpuinfo.get_cpu_info()
    flags = info.get('flags', [])

    if 'avx2' in flags:
        return True

    return False


def get_github_release(repo: str) -> dict:
    api_url = f"https://api.github.com/repos/{repo}/releases/latest"

    try:
        response = requests.get(api_url, timeout=DEFAULT_REQUEST_TIMEOUT)
        response.raise_for_status()
        release_data = response.json()
        assets = release_data.get('assets', [])
        return {asset['name']: asset['browser_download_url'] for asset in assets}
    except (json.JSONDecodeError, requests.RequestException) as e:
        logging.error("Failed to fetch release data from GitHub: %s", e)
    return {}


# This is necessary to keep due to the error that occurs when using py7zr.
# Error: (b'\x03\x03\x01\x1b', 'BCJ2 filter is not supported by py7zr.
# Please consider to contribute to XZ/liblzma project and help Python core team implementing it.
# Or please use another tool to extract it.')

def download_7z(zip_tool: str) -> None:
    if not os.path.exists(zip_tool):
        print("Downloading 7z...")
        download_file('https://7-zip.org/a/7zr.exe', zip_tool)


def download_mpv(dep_path: str = None, appdata_path: str = None, update: bool = False):  # pylint: disable=too-many-branches, too-many-locals, too-many-branches
    if update:
        print("Updating MPV...")

    if sys.platform == 'darwin':
        if shutil.which("brew"):
            if update:
                print("Updating MPV using Homebrew...")
                subprocess.run(["brew", "update"], check=True,
                               stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                subprocess.run(["brew", "upgrade", "--formula", "mpv"], check=True,
                               stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                return
            if not shutil.which("mpv"):
                print("Installing MPV using Homebrew...")
                subprocess.run(["brew", "update"], check=True,
                               stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                subprocess.run(["brew", "install", "--formula", "mpv"], check=True,
                               stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                return
        return

    if sys.platform == 'linux' and update:
        print("Not implemented.")
        return

    appdata_path = appdata_path or os.path.join(
        os.environ['USERPROFILE'], 'AppData', 'Roaming', 'aniworld'
    )
    dep_path = dep_path or os.path.join(appdata_path, "mpv")
    if update is True:
        if os.path.exists(dep_path):
            shutil.rmtree(dep_path)
    os.makedirs(dep_path, exist_ok=True)

    executable_path = os.path.join(dep_path, 'mpv.exe')
    zip_path = os.path.join(dep_path, 'mpv.7z')
    zip_tool = os.path.join(appdata_path, "7z", "7zr.exe")
    os.makedirs(os.path.dirname(zip_tool), exist_ok=True)

    if os.path.exists(executable_path):
        return

    direct_links = get_github_release("shinchiro/mpv-winbuild-cmake")
    avx2_supported = check_avx2_support()
    pattern = (
        r'mpv-x86_64-v3-\d{8}-git-[a-f0-9]{7}\.7z'
        if avx2_supported
        else r'mpv-x86_64-\d{8}-git-[a-f0-9]{7}\.7z'
    )
    logging.debug("Downloading MPV using pattern: %s", pattern)
    direct_link = next(
        (link for name, link in direct_links.items() if re.match(pattern, name)),
        None
    )

    if not direct_link:
        logging.error(
            "No suitable MPV download link found. Please download manually.")
        return

    if not os.path.exists(zip_path):
        logging.debug("Downloading MPV from %s to %s", direct_link, zip_path)
        print(
            f"Downloading MPV ({'without' if not avx2_supported else 'with'} AVX2)...")
        download_file(direct_link, zip_path)

    download_7z(zip_tool)

    logging.debug("Extracting MPV to %s", dep_path)
    try:
        subprocess.run(
            [zip_tool, "x", zip_path],
            check=True,
            cwd=dep_path,
            stdout=subprocess.DEVNULL
        )
    except (subprocess.CalledProcessError, FileNotFoundError, OSError,
            subprocess.SubprocessError) as e:
        logging.error("Failed to extract files: %s", e)

    logging.debug("Adding MPV path to environment: %s", dep_path)
    os.environ["PATH"] += os.pathsep + dep_path

    if os.path.exists(zip_path):
        os.remove(zip_path)


def download_syncplay(dep_path: str = None, appdata_path: str = None, update: bool = False):
    if update:
        print("Updating Syncplay...")

    if sys.platform == 'darwin' and update:
        if shutil.which("brew"):
            print("Updating Syncplay using Homebrew...")
            subprocess.run(["brew", "update"], check=True,
                           stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            subprocess.run(
                ["brew", "upgrade", "--cask", "syncplay"],
                check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
        return

    if sys.platform == 'linux' and update:
        print("Not implemented.")
        return

    appdata_path = appdata_path or os.path.join(
        os.environ['USERPROFILE'], 'AppData', 'Roaming', 'aniworld')
    dep_path = dep_path or os.path.join(appdata_path, "syncplay")
    if update is True:
        if os.path.exists(dep_path):
            shutil.rmtree(dep_path)
    os.makedirs(dep_path, exist_ok=True)

    executable_path = os.path.join(dep_path, 'SyncplayConsole.exe')
    zip_path = os.path.join(dep_path, 'syncplay.zip')

    if os.path.exists(executable_path):
        return

    direct_links = get_github_release("Syncplay/syncplay")
    direct_link = next(
        (link for name, link in direct_links.items()
         if re.match(r'Syncplay_\d+\.\d+\.\d+_Portable\.zip', name)),
        None
    )

    if not os.path.exists(executable_path):
        print("Downloading Syncplay...")
        download_file(direct_link, zip_path)

    logging.debug("Extracting Syncplay to %s", dep_path)
    try:
        subprocess.run(
            ["tar", "-xf", zip_path],
            check=True,
            cwd=dep_path
        )
    except subprocess.CalledProcessError as e:
        logging.error("Failed to extract files: %s", e)
    except FileNotFoundError:
        logging.error("7zr.exe not found at the specified path.")
    except subprocess.SubprocessError as e:
        logging.error("An error occurred: %s", e)

    if os.path.exists(zip_path):
        os.remove(zip_path)


def download_file(url: str, path: str):
    try:
        response = requests.get(
            url, stream=True, allow_redirects=True, timeout=DEFAULT_REQUEST_TIMEOUT)
        total_size = int(response.headers.get('content-length', 0))
        block_size = 1024
        t = tqdm(total=total_size, unit='B', unit_scale=True)
        with open(path, 'wb') as f:
            for data in response.iter_content(block_size):
                t.update(len(data))
                f.write(data)
        t.close()
    except requests.RequestException as e:
        logging.error("Failed to download: %s", e)


def get_season_episode_count(slug) -> dict:
    base_url = f"{ANIWORLD_TO}/anime/stream/{slug}/"
    response = requests.get(base_url, timeout=DEFAULT_REQUEST_TIMEOUT)
    soup = BeautifulSoup(response.content, 'html.parser')

    season_meta = soup.find('meta', itemprop='numberOfSeasons')
    number_of_seasons = int(season_meta['content']) if season_meta else 0

    episode_counts = {}

    for season in range(1, number_of_seasons + 1):
        season_url = f"{base_url}staffel-{season}"
        response = requests.get(season_url, timeout=DEFAULT_REQUEST_TIMEOUT)
        soup = BeautifulSoup(response.content, 'html.parser')

        episode_links = soup.find_all('a', href=True)
        unique_links = set(
            link['href']
            for link in episode_links
            if f"staffel-{season}/episode-" in link['href']
        )

        episode_counts[season] = len(unique_links)

    return episode_counts


def get_movie_episode_count(slug) -> int:
    movie_page_url = f"{ANIWORLD_TO}/anime/stream/{slug}/filme"
    response = requests.get(
        movie_page_url, timeout=DEFAULT_REQUEST_TIMEOUT)

    parsed_html = BeautifulSoup(response.content, 'html.parser')
    movie_indices = []

    movie_index = 1
    while True:
        expected_subpath = f"{slug}/filme/film-{movie_index}"

        matching_links = [link['href'] for link in parsed_html.find_all(
            'a', href=True) if expected_subpath in link['href']]

        if matching_links:
            movie_indices.append(movie_index)
            movie_index += 1
        else:
            break

    # has_movies = bool(movie_indices)
    return max(movie_indices) if movie_indices else 0


def generate_links(urls, arguments):  # pylint: disable=too-many-locals, too-many-branches
    """
    Example usage:
    seasons = {1: 12, 2: 13, 3: 4}
    base_url = [
        "https://aniworld.to/anime/stream/food-wars-shokugeki-no-sma/staffel-1/episode-1",
        "https://aniworld.to/anime/stream/food-wars-shokugeki-no-sma/staffel-2",
        "https://aniworld.to/anime/stream/overlord"
    ]
    result = generate_links(base_url)

    for url in result:
        print(url)
    """

    unique_links = set()

    # TODO: save season data for every anime and use this cached season info in the
    # for loop instead of generating the season_info each time
    for base_url in urls:
        seasons_info = {}
        movies_info = 0
        parts = base_url.split('/')

        if ("anime" in parts and not "episode" in base_url or
                "anime" in parts and arguments.keep_watching):
            series_slug_index = parts.index("stream") + 1
            series_slug = parts[series_slug_index]
            seasons_info = get_season_episode_count(
                slug=series_slug
            )
            movies_info = get_movie_episode_count(
                slug=series_slug
            )
        else:
            unique_links.add(base_url)
            continue

        # print(seasons_info)

        if base_url.endswith("/"):
            base_url = base_url[:-1]

        parts = base_url.split("/")

        if arguments.keep_watching:
            season_start = 1
            episode_start = 1
            movie_start = 1
            season_match = re.search(r"staffel-(\d+)", base_url)
            episode_match = re.search(r"episode-(\d+)", base_url)
            movie_match = re.search(r"film-(\d+)", base_url)

            if season_match:
                season_start = int(season_match.group(1))

            if episode_match:
                episode_start = int(episode_match.group(1))

            if movie_match:
                movie_start = int(movie_match.group(1))

            raw_url = "/".join(base_url.split("/")[:6])

            if "film" not in base_url:
                for season in range(season_start, len(seasons_info) + 1):
                    season_url = f"{raw_url}/staffel-{season}/"
                    for episode in range(episode_start, seasons_info[season] + 1):
                        unique_links.add(f"{season_url}episode-{episode}")
                    episode_start = 1
                continue

            for episode in range(movie_start, movies_info + 1):
                unique_links.add(f"{raw_url}/filme/film-{episode}")
            continue

        if "staffel" not in base_url and "episode" not in base_url and not "film" in base_url:
            for season, episodes in seasons_info.items():
                season_url = f"{base_url}/staffel-{season}/"
                for episode in range(1, episodes + 1):
                    unique_links.add(f"{season_url}episode-{episode}")
            continue

        if "staffel" in base_url and "episode" not in base_url:
            season = int(parts[-1].split("-")[-1])
            if season in seasons_info:
                for episode in range(1, seasons_info[season] + 1):
                    unique_links.add(f"{base_url}/episode-{episode}")
            continue

        if "filme" in base_url and "film-" not in base_url:
            for episode in range(1, movies_info + 1):
                unique_links.add(f"{base_url}/film-{episode}")
            continue

        unique_links.add(base_url)

    def natural_sort_key(link_url):
        return [int(text) if text.isdigit() else text for text in re.split(r'(\d+)', link_url)]

    return sorted(unique_links, key=natural_sort_key)


def remove_anime4k():
    anime4k_shader_path = os.path.join(MPV_DIRECTORY, "shaders")
    anime4k_input_conf_path = os.path.join(MPV_DIRECTORY, "input.conf")
    anime4k_mpv_conf_path = os.path.join(MPV_DIRECTORY, "mpv.conf")

    if os.path.exists(anime4k_shader_path):
        print(f"Removing: {anime4k_shader_path}")
        shutil.rmtree(anime4k_shader_path)

    if os.path.exists(anime4k_input_conf_path):
        print(f"Removing: {anime4k_input_conf_path}")
        os.remove(anime4k_input_conf_path)

    if os.path.exists(anime4k_mpv_conf_path):
        print(f"Removing: {anime4k_mpv_conf_path}")
        os.remove(anime4k_mpv_conf_path)


def remove_mpv_scripts():
    scripts = ["aniskip.lua", "autoexit.lua", "autostart.lua"]

    for script in scripts:
        script_path = os.path.join(MPV_DIRECTORY, "scripts", script)
        if os.path.exists(script_path):
            print(f"Removing: {script_path}")
            os.remove(script_path)


def copy_file_if_different(source_path, destination_path):
    if os.path.exists(destination_path):
        with open(source_path, 'r', encoding="utf-8") as source_file:
            source_content = source_file.read()

        with open(destination_path, 'r', encoding="utf-8") as destination_file:
            destination_content = destination_file.read()

        if source_content != destination_content:
            logging.debug(
                "Content differs, overwriting %s", os.path.basename(
                    destination_path
                )
            )
            shutil.copy(source_path, destination_path)
        else:
            logging.debug(
                "%s already exists and is identical, no overwrite needed",
                os.path.basename(destination_path)
            )
    else:
        logging.debug(
            "Copying %s to %s",
            os.path.basename(source_path),
            os.path.dirname(destination_path)
        )
        shutil.copy(source_path, destination_path)


def setup_autostart():
    script_directory = os.path.dirname(
        os.path.dirname(os.path.abspath(__file__))
    )

    mpv_scripts_directory = MPV_SCRIPTS_DIRECTORY

    if not os.path.exists(mpv_scripts_directory):
        os.makedirs(mpv_scripts_directory)

    autostart_source_path = os.path.join(
        script_directory, 'aniskip', 'scripts', 'autostart.lua'
    )

    autostart_destination_path = os.path.join(
        mpv_scripts_directory, 'autostart.lua'
    )

    logging.debug("Copying %s to %s if needed.",
                  autostart_source_path, autostart_destination_path)
    copy_file_if_different(autostart_source_path, autostart_destination_path)


def setup_autoexit():
    logging.debug("Copying autoexit.lua to mpv script directory")
    script_directory = os.path.dirname(
        os.path.dirname(os.path.abspath(__file__)))
    mpv_scripts_directory = MPV_SCRIPTS_DIRECTORY

    if not os.path.exists(mpv_scripts_directory):
        os.makedirs(mpv_scripts_directory)

    autoexit_source_path = os.path.join(
        script_directory, 'aniskip', 'scripts', 'autoexit.lua')
    autoexit_destination_path = os.path.join(
        mpv_scripts_directory, 'autoexit.lua')

    copy_file_if_different(autoexit_source_path, autoexit_destination_path)


if __name__ == "__main__":
    print(check_avx2_support())



import argparse
import importlib
import json
import os
import logging
import platform
import random
import subprocess
import shutil
import sys

import requests

from aniworld.common import download_mpv, download_syncplay, remove_anime4k, remove_mpv_scripts
from aniworld.extractors import get_direct_link_from_hanime
from aniworld.anime4k import download_anime4k
from aniworld import config


class CaseInsensitiveChoices:  # pylint: disable=too-few-public-methods
    def __init__(self, choices):
        self.choices = choices
        self.normalized = {c.lower(): c for c in choices}

    def __call__(self, value):
        key = value.lower()
        if key in self.normalized:
            return self.normalized[key]
        raise argparse.ArgumentTypeError(
            f"invalid choice: {value} (choose from {', '.join(self.choices)})")


def get_random_anime_slug(genre) -> str:
    url = f'{config.ANIWORLD_TO}/ajax/randomGeneratorSeries'

    data = {
        'productionStart': 'all',
        'productionEnd': 'all', 'genres[]': genre
    }

    headers = {'User-Agent': config.RANDOM_USER_AGENT}

    try:
        response = requests.post(
            url, data=data, headers=headers, timeout=config.DEFAULT_REQUEST_TIMEOUT
        )

        response.raise_for_status()
        anime_list = response.json()

        if not anime_list:
            logging.warning("No anime found for genre: %s", genre)
            return None

        random_anime = random.choice(anime_list)
        logging.debug("Selected Anime: %s", random_anime)

        return random_anime.get('link')

    except requests.RequestException as e:
        logging.error("Request failed: %s", e)
    except (json.JSONDecodeError, KeyError, TypeError) as e:
        logging.error("Error processing response: %s", e)

    return None


def parse_arguments() -> argparse.Namespace:  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    parser = argparse.ArgumentParser(
        description="Parse command-line arguments for anime streaming, "
                    "downloading, and playback management."
    )

    # General options
    general_opts = parser.add_argument_group('General Options')
    general_opts.add_argument(
        '-d', '--debug',
        action='store_true',
        help='Enable debug mode for detailed logs.'
    )
    general_opts.add_argument(
        '-U', '--update',
        type=str,
        choices=['mpv', 'yt-dlp', 'syncplay', 'all'],
        help='Update specified tools (mpv, yt-dlp, syncplay, or all).'
    )
    general_opts.add_argument(
        '-u', '--uninstall',
        action='store_true',
        help='Perform self-uninstallation.'
    )
    general_opts.add_argument(
        '-v', '--version',
        action='store_true',
        help='Display version information.'
    )

    # Search options
    search_opts = parser.add_argument_group('Search Options')
    search_opts.add_argument(
        '-s', '--slug',
        type=str,
        help='Specify a search slug (e.g., demon-slayer-kimetsu-no-yaiba).'
    )

    # Episode options
    episode_opts = parser.add_argument_group('Episode Options')
    episode_opts.add_argument(
        '-e', '--episode',
        type=str,
        nargs='+',
        help='Specify one or more episode URLs.'
    )
    episode_opts.add_argument(
        '-f', '--episode-file',
        type=str,
        help='Provide a file containing episode URLs.'
    )
    episode_opts.add_argument(
        '-lf', '--local-episodes',
        type=str,
        nargs='+',
        help='Use local MP4 files for episodes instead of URLs.'
    )
    episode_opts.add_argument(
        '-pl', '--provider-link',
        type=str,
        nargs='+',
        help='Specify one or more provider episode urls.'
    )

    # Action options
    action_opts = parser.add_argument_group('Action Options')
    action_opts.add_argument(
        '-a', '--action',
        type=CaseInsensitiveChoices(['Watch', 'Download', 'Syncplay']),
        default=config.DEFAULT_ACTION,
        help='Specify the action to perform.'
    )
    action_opts.add_argument(
        '-o', '--output-dir',
        type=str,
        default=config.DEFAULT_DOWNLOAD_PATH,
        help='Set the download directory (e.g., /path/to/downloads).'
    )
    action_opts.add_argument(
        '-L', '--language',
        type=CaseInsensitiveChoices(
            ['German Dub', 'English Sub', 'German Sub']
        ),
        default=config.DEFAULT_LANGUAGE,
        help='Specify the language for playback or download.'
    )
    action_opts.add_argument(
        '-p', '--provider',
        type=CaseInsensitiveChoices(config.SUPPORTED_PROVIDERS),
        help='Specify the preferred provider.'
    )

    # Anime4K options
    anime4k_opts = parser.add_argument_group('Anime4K Options')
    anime4k_opts.add_argument(
        '-A', '--anime4k',
        type=CaseInsensitiveChoices(['High', 'Low', 'Remove']),
        help='Set Anime4K mode (High, Low, or Remove for performance tuning).'
    )

    # Syncplay options
    syncplay_opts = parser.add_argument_group('Syncplay Options')
    syncplay_opts.add_argument(
        '-sH', '--hostname',
        type=str,
        help='Set the Syncplay server hostname.'
    )
    syncplay_opts.add_argument(
        '-sU', '--username',
        type=str,
        help='Set the Syncplay username.'
    )
    syncplay_opts.add_argument(
        '-sR', '--room',
        type=str,
        help='Specify the Syncplay room name.'
    )
    syncplay_opts.add_argument(
        '-sP', '--password',
        type=str,
        nargs='+',
        help='Set the Syncplay room password.'
    )

    # Miscellaneous options
    misc_opts = parser.add_argument_group('Miscellaneous Options')
    misc_opts.add_argument(
        '-k', '--aniskip',
        action='store_true',
        help='Skip anime intros and outros using Aniskip.'
    )
    misc_opts.add_argument(
        '-K', '--keep-watching',
        action='store_true',
        help='Automatically continue to the next episodes after the selected one.'
    )
    misc_opts.add_argument(
        '-r', '--random-anime',
        type=str,
        nargs='*',
        help='Play a random anime (default genre is "all", e.g., Drama).\n'
             f'All genres can be found here: "{config.ANIWORLD_TO}/random"'
    )
    misc_opts.add_argument(
        '-D', '--only-direct-link',
        action='store_true',
        help='Output only the direct streaming link.'
    )
    misc_opts.add_argument(
        '-C', '--only-command',
        action='store_true',
        help='Output only the execution command.'
    )

    args = parser.parse_args()

    if args.uninstall:
        print(f"Removing: {config.DEFAULT_APPDATA_PATH}")
        if os.path.exists(config.DEFAULT_APPDATA_PATH):
            shutil.rmtree(config.DEFAULT_APPDATA_PATH)

        remove_anime4k()
        remove_mpv_scripts()

        if sys.platform.startswith('win'):
            command = "timeout 3 >nul & pip uninstall -y aniworld"
        else:
            command = "pip uninstall -y aniworld"

        print("pip uninstall -y aniworld")
        subprocess.Popen(  # pylint: disable=consider-using-with
            command,
            shell=True,
            creationflags=subprocess.CREATE_NEW_CONSOLE
        )

        sys.exit()

    if args.version:
        cowsay = fR"""
_____________________________
< AniWorld-Downloader v.{config.VERSION} >
-----------------------------
        \   ^__^
         \  (oo)\_______
            (__)\       )\/\
                ||----w |
                ||     ||
"""
        if not config.IS_NEWEST_VERSION:
            cowsay += (
                f"\nYour version is outdated.\n"
                f"Please update to the latest version (v.{config.LATEST_VERSION})."
            )
        else:
            cowsay += "\nYou are on the latest version."
        print(cowsay.strip())
        sys.exit()

    if args.anime4k:
        download_anime4k(args.anime4k)

    if args.provider_link:  # pylint: disable=too-many-nested-blocks
        invalid_links = [
            link for link in args.provider_link if not link.startswith("http")
        ]
        if invalid_links:
            print(f"Invalid provider episode URLs: {', '.join(invalid_links)}")
            sys.exit(1)
        for link in args.provider_link:
            if link.startswith("https://hanime.tv/videos/"):
                get_direct_link_from_hanime(link)

        args.provider_link = [link for link in args.provider_link if not link.startswith(
            "https://hanime.tv/videos/")]

        if not args.provider_link:
            sys.exit()

        if not args.provider:
            print("Provider must be specified when using provider links.")
            sys.exit(1)

        logging.info("Using provider: %s", args.provider)

        if args.provider in config.SUPPORTED_PROVIDERS:
            module = importlib.import_module("aniworld.extractors")
            func = getattr(
                module, f"get_direct_link_from_{args.provider.lower()}"
            )

            for provider_episode in args.provider_link:
                direct_link = f'"{func(provider_episode)}"'

                if args.provider in config.PROVIDER_HEADERS:
                    if config.PROVIDER_HEADERS.get(args.provider):
                        action = (
                            config.YTDLP_PATH if args.action == "Download" else
                            config.MPV_PATH if args.action == "Watch" else
                            config.SYNCPLAY_PATH if args.action == "Syncplay" else
                            None
                        )

                        if action:
                            header = (
                                "--add-header"
                                if args.action == "Download"
                                else "--http-header-fields"
                            )
                            direct_link = (
                                f"{action} {direct_link} "
                                f"{header}='{config.PROVIDER_HEADERS[args.provider]}'"
                            )
                        else:
                            raise ValueError("Invalid action.")

                print(f"-> {provider_episode}")
                print(direct_link)
                print("-" * 40)

            sys.exit()

    if args.update:
        def update_yt_dlp():
            logging.info("Upgrading yt-dlp...")
            yt_dlp_update_command = ["pip", "install", "-U", "yt-dlp"]

            logging.debug("Running Command: %s", yt_dlp_update_command)
            subprocess.run(
                yt_dlp_update_command,
                check=False,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )

        def update_all():
            logging.info("Updating all tools...")
            download_mpv(update=True)
            update_yt_dlp()
            download_syncplay(update=True)

        update_actions = {
            "mpv": lambda: download_mpv(update=True),
            "yt-dlp": update_yt_dlp,
            "syncplay": lambda: download_syncplay(update=True),
            "all": update_all
        }

        action = update_actions.get(args.update)
        if action:
            action()
        else:
            logging.error("Invalid update option provided.")

    if args.random_anime is not None:
        args.slug = get_random_anime_slug(
            args.random_anime if len(args.random_anime) != 0 else "all")

    if args.provider is None:
        config.USES_DEFAULT_PROVIDER = True  # pylint: disable=global-statement

        args.provider = (
            config.DEFAULT_PROVIDER_DOWNLOAD
            if args.action == "Download" else config.DEFAULT_PROVIDER_WATCH
        )

    if args.debug is True:
        def open_terminal_with_command(command):
            if os.environ.get('DISPLAY'):
                terminal_emulators = [
                    ('gnome-terminal', ['gnome-terminal', '--',
                     'bash', '-c', f'{command}; exec bash']),
                    ('xterm', ['xterm', '-hold', '-e', command]),
                    ('konsole', ['konsole', '--hold', '-e', command])
                ]

                for terminal, cmd in terminal_emulators:
                    try:
                        subprocess.Popen(  # pylint: disable=consider-using-with
                            cmd
                        )
                        return
                    except FileNotFoundError:
                        logging.debug(
                            "%s not found, trying next option.", terminal
                        )
                    except subprocess.SubprocessError as e:
                        logging.error(
                            "Error opening terminal with %s: %s", terminal, e)

                logging.error(
                    "No supported terminal emulator found. "
                    "Install gnome-terminal, xterm, or konsole.")
            print("It looks like you are on a headless machine! "
                  "For advanced log look in your temp folder!")
            return

        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug("=============================================")
        logging.debug(
            "   Welcome to AniWorld Downloader v.%s!   ", config.VERSION)
        logging.debug("=============================================\n")

        system = platform.system()

        if system == "Darwin":
            try:
                darwin_open_debug_log = [
                    "osascript", "-e",
                    'tell application "Terminal" to do script "trap exit SIGINT; '
                    'tail -f -n +1 $TMPDIR/aniworld.log" activate'
                ]
                logging.debug("Running Command: %s", darwin_open_debug_log)
                subprocess.run(darwin_open_debug_log, check=True)
            except subprocess.CalledProcessError as e:
                logging.error("Failed to start tailing the log file: %s", e)
        elif system == "Windows":
            try:
                windows_open_debug_log = (
                    "start cmd /c \"powershell -NoExit -c "
                    "Get-Content -Wait \"$env:TEMP\\aniworld.log\"\""
                )
                logging.debug("Running Command: %s", windows_open_debug_log)
                subprocess.run(windows_open_debug_log, shell=True, check=True)
            except subprocess.CalledProcessError as e:
                logging.error("Failed to start tailing the log file: %s", e)
        elif system == "Linux":
            open_terminal_with_command('tail -f -n +1 /tmp/aniworld.log')

    return args


arguments = parse_arguments()

if __name__ == "__main__":
    pass



from aniworld.entry import aniworld
from aniworld.config import VERSION, IS_NEWEST_VERSION


def main():
    print(
        f"\033]0;AniWorld-Downloader v.{VERSION}"
        f"{' (Update Available)' if not IS_NEWEST_VERSION else ''}\007",
        end='',
        flush=True
    )

    aniworld()


if __name__ == "__main__":
    main()



import json
import html
import webbrowser
from urllib.parse import quote
import logging

import curses
import requests

from aniworld.ascii_art import display_ascii_art
from aniworld.config import DEFAULT_REQUEST_TIMEOUT, ANIWORLD_TO


def search_anime(keyword: str = None) -> str:
    print(display_ascii_art())
    if not keyword:
        logging.debug("Prompting user for search.")
        keyword = input("Search for a series: ").strip()
        if keyword.strip().lower() == "boku no piko":
            raise ValueError("Really? This is not on AniWorld...")

    search_url = f"{ANIWORLD_TO}/ajax/seriesSearch?keyword={quote(keyword)}"
    anime_list = fetch_anime_list(search_url)

    if len(anime_list) == 1:
        return anime_list[0].get("link", None)

    if not anime_list:
        raise ValueError("Could not get valid anime")

    return curses.wrapper(show_menu, anime_list)


def fetch_anime_list(url: str) -> list:
    try:
        response = requests.get(url, timeout=DEFAULT_REQUEST_TIMEOUT)
        response.raise_for_status()
        decoded_data = json.loads(html.unescape(response.text))
        if isinstance(decoded_data, list):
            return decoded_data
        return []
    except (requests.RequestException, json.JSONDecodeError) as exc:
        raise ValueError("Could not get valid anime: ") from exc


def show_menu(stdscr: curses.window, options: list) -> str:
    current_row = 0
    konami_code = ['UP', 'UP', 'DOWN', 'DOWN',
                   'LEFT', 'RIGHT', 'LEFT', 'RIGHT', 'b', 'a']
    entered_keys = []
    key_map = {
        curses.KEY_UP: 'UP',
        curses.KEY_DOWN: 'DOWN',
        curses.KEY_LEFT: 'LEFT',
        curses.KEY_RIGHT: 'RIGHT',
        ord('b'): 'b',
        ord('a'): 'a'
    }

    try:
        while True:
            stdscr.clear()
            for idx, anime in enumerate(options):
                name = anime.get('name', 'No Name')
                year = anime.get('productionYear', 'Unknown Year')
                highlight = curses.A_REVERSE if idx == current_row else 0
                stdscr.attron(highlight)
                stdscr.addstr(idx, 0, f"{name} {year}")
                stdscr.attroff(highlight)

            stdscr.refresh()
            key = stdscr.getch()

            if key in key_map:
                entered_keys.append(key_map[key])
                if len(entered_keys) > len(konami_code):
                    entered_keys.pop(0)
                if entered_keys == konami_code:
                    webbrowser.open(
                        'https://www.youtube.com/watch?v=PDJLvF1dUek')
                    entered_keys.clear()
            else:
                entered_keys.clear()

            if key == curses.KEY_DOWN:
                current_row = (current_row + 1) % len(options)
            elif key == curses.KEY_UP:
                current_row = (current_row - 1 + len(options)) % len(options)
            elif key == ord('\n'):
                return options[current_row].get('link', 'No Link')
            elif key == ord('q'):
                break
    except curses.error:
        pass

    return None


if __name__ == '__main__':
    print(search_anime())



import getpass
import subprocess
import logging
import hashlib

from aniworld.models import Anime
from aniworld.config import MPV_PATH, PROVIDER_HEADERS, SYNCPLAY_PATH, INVALID_PATH_CHARS
from aniworld.common import download_mpv, download_syncplay, setup_autostart, setup_autoexit
from aniworld.aniskip import aniskip
from aniworld.parser import arguments


def _get_syncplay_username():
    return arguments.username or getpass.getuser()


def _get_syncplay_hostname():
    return arguments.hostname or "syncplay.pl:8997"


def _get_syncplay_room(title):
    if arguments.room:
        return arguments.room

    room = title
    password = arguments.password

    if password:
        room += f":{password}"

    room_hash = hashlib.sha256(room.encode('utf-8')).hexdigest()

    return f"AniWorld_Downloader.{room_hash}"


def _append_password(command, title):
    password = arguments.password

    if not password:
        return command

    password = f"{arguments.password}:{title}"
    password_hash = hashlib.sha256(password.encode('utf-8')).hexdigest()

    command.insert(9, "--password")
    command.insert(10, password_hash)
    return command


def _execute_command(command):
    if arguments.only_command:
        print("\n" + " ".join(str(item) for item in command))
        return
    try:
        logging.debug("Running Command:\n%s", command)
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        logging.error(
            "Error running command: %s\nCommand: %s",
            e, ' '.join(
                str(item) if item is not None else '' for item in command)
        )


def _build_syncplay_command(source, title=None, headers=None, aniskip_data=None,
                            anime=None, media_title=None):
    command = [
        SYNCPLAY_PATH,
        "--no-gui",
        "--no-store",
        "--host", _get_syncplay_hostname(),
        "--room", _get_syncplay_room(title=title),
        "--name", _get_syncplay_username(),
        "--player-path", MPV_PATH,
        source,
        "--",
        "--fs"
    ]

    if media_title:
        command.append(f'--force-media-title="{media_title}"')

    command = _append_password(command, title)

    if headers:
        if anime.provider != "Luluvdo":
            for header in headers:
                command.append(f"--http-header-fields={header}")
        else:
            command.append(f"--http-header-fields={headers[0]}")

    if aniskip_data:
        command.extend(aniskip_data.split()[:2])

    return command


def _process_anime_episodes(anime):
    for episode in anime:
        episode_title = f"{anime.title} - S{episode.season}E{episode.episode} - ({anime.language}):"
        direct_link = episode.get_direct_link()

        if not direct_link:
            logging.warning(f"Something went wrong with \"{episode_title}\".")
            continue

        if arguments.only_direct_link:
            print(
                episode_title
            )
            print(f"{direct_link}\n")
            continue

        sanitized_anime_title = ''.join(
            char for char in anime.title if char not in INVALID_PATH_CHARS
        )

        if episode.season == 0:
            media_title = (
                f"{sanitized_anime_title} - "
                f"Movie {episode.episode:02} - "
                f"({anime.language})"
            )
        else:
            media_title = (
                f"{sanitized_anime_title} - "
                f"S{episode.season:02}E{episode.episode:03} - "
                f"({anime.language})"
            )

        command = _build_syncplay_command(
            direct_link,
            episode.title_german,
            PROVIDER_HEADERS.get(anime.provider),
            aniskip(anime.title, episode.episode,
                    episode.season, episode.season_episode_count[episode.season])
            if anime.aniskip else None,
            anime,
            media_title
        )
        _execute_command(command)


def _process_local_files():
    for file in arguments.local_episodes:
        command = _build_syncplay_command(file)
        _execute_command(command)


def syncplay(anime: Anime = None):
    download_mpv()
    download_syncplay()

    setup_autostart()
    setup_autoexit()

    if anime is None:
        _process_local_files()
    else:
        _process_anime_episodes(anime)


if __name__ == '__main__':
    download_syncplay()



import sys

from aniworld.models import Anime
from aniworld.action import watch, download, syncplay


def execute(anime_list: list[Anime]):
    for anime in anime_list:
        try:
            if anime.action == "Watch":
                watch(anime)
            elif anime.action == "Download":
                download(anime)
            elif anime.action == "Syncplay":
                syncplay(anime)
            else:
                raise ValueError("Invalid action specified for anime: {anime}")
        except AttributeError:
            sys.exit()



import subprocess
import logging

from aniworld.aniskip import aniskip
from aniworld.common import download_mpv
from aniworld.config import MPV_PATH, PROVIDER_HEADERS, INVALID_PATH_CHARS
from aniworld.models import Anime
from aniworld.parser import arguments


def _build_watch_command(source, media_title=None, headers=None, aniskip_data=None, anime=None):
    command = [MPV_PATH, source, "--fs", "--quiet"]
    if media_title:
        command.append(f'--force-media-title="{media_title}"')
    if headers:
        if anime.provider != "Luluvdo":
            for header in headers:
                command.append(f"--http-header-fields={header}")
        else:
            command.append(f"--http-header-fields={headers[0]}")
    if aniskip_data:
        command.extend(aniskip_data.split()[:2])
    return command


def _print_or_run(title, command):
    if arguments.only_command:
        print(f"\n{title}:")
        print(" ".join(str(item) for item in command if item is not None))
        return
    try:
        logging.debug("Running Command:\n%s", command)
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        logging.error(
            "Error running command: %s\nCommand: %s",
            e, ' '.join(
                str(item) if item is not None else '' for item in command)
        )


def _process_local_files():
    for file in arguments.local_episodes:
        command = _build_watch_command(file)
        _print_or_run(file, command)


def _process_anime_episodes(anime):
    for episode in anime:
        episode_title = f"{anime.title} - S{episode.season}E{episode.episode} - ({anime.language}):"
        direct_link = episode.get_direct_link()

        if not direct_link:
            logging.warning(f"Something went wrong with \"{episode_title}\".")
            continue

        if arguments.only_direct_link:
            print(episode_title)
            print(f"{direct_link}\n")
            continue

        sanitized_anime_title = ''.join(
            char for char in anime.title if char not in INVALID_PATH_CHARS
        )

        if episode.season == 0:
            media_title = (
                f"{sanitized_anime_title} - "
                f"Movie {episode.episode:02} - "
                f"({anime.language})"
            )
        else:
            media_title = (
                f"{sanitized_anime_title} - "
                f"S{episode.season:02}E{episode.episode:03} - "
                f"({anime.language})"
            )

        title = _generate_episode_title(anime, episode)
        command = _build_watch_command(
            direct_link,
            media_title,
            PROVIDER_HEADERS.get(anime.provider),
            aniskip(anime.title, episode.episode,
                    episode.season, episode.season_episode_count[episode.season])
            if anime.aniskip else None,
            anime
        )
        _print_or_run(title, command)


def _generate_episode_title(anime, episode):
    if episode.has_movies and episode.season not in episode.season_episode_count:
        return f"{anime.title} - Movie {episode.episode} - {episode.title_german}"
    return f"{anime.title} - S{episode.season}E{episode.episode} - {episode.title_german}"


def watch(anime: Anime = None):
    download_mpv()

    if anime is None:
        _process_local_files()
    else:
        _process_anime_episodes(anime)



import re
import random
import time

import requests

from aniworld.config import RANDOM_USER_AGENT, DEFAULT_REQUEST_TIMEOUT


def get_direct_link_from_doodstream(embeded_doodstream_link):
    headers = {
        'User-Agent': RANDOM_USER_AGENT,
        'Referer': 'https://dood.li/'
    }

    def extract_data(pattern, content):
        match = re.search(pattern, content)
        return match.group(1) if match else None

    def generate_random_string(length=10):
        characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
        return ''.join(random.choice(characters) for _ in range(length))

    response = requests.get(
        embeded_doodstream_link,
        headers=headers,
        timeout=DEFAULT_REQUEST_TIMEOUT,
        verify=False
    )
    response.raise_for_status()

    pass_md5_pattern = r"\$\.get\('([^']*\/pass_md5\/[^']*)'"
    pass_md5_url = extract_data(pass_md5_pattern, response.text)
    if not pass_md5_url:
        raise ValueError(
            f'pass_md5 URL not found using {embeded_doodstream_link}.')

    full_md5_url = f"https://dood.li{pass_md5_url}"

    token_pattern = r"token=([a-zA-Z0-9]+)"
    token = extract_data(token_pattern, response.text)
    if not token:
        raise ValueError(f'Token not found using {embeded_doodstream_link}.')

    md5_response = requests.get(
        full_md5_url, headers=headers, timeout=DEFAULT_REQUEST_TIMEOUT, verify=False)
    md5_response.raise_for_status()
    video_base_url = md5_response.text.strip()

    random_string = generate_random_string(10)
    expiry = int(time.time())

    direct_link = f"{video_base_url}{random_string}?token={token}&expiry={expiry}"
    # print(direct_link)

    return direct_link


if __name__ == '__main__':
    link = input("Enter Doodstream Link: ")
    print(get_direct_link_from_doodstream(embeded_doodstream_link=link))



import os
import curses
import sys
import logging
import urllib3
import npyscreen

from aniworld.models import Anime, Episode
from aniworld.config import (
    VERSION,
    SUPPORTED_PROVIDERS,
    DEFAULT_PROVIDER_DOWNLOAD,
    DEFAULT_PROVIDER_WATCH,
    USES_DEFAULT_PROVIDER,
    IS_NEWEST_VERSION,
    ANIWORLD_TO
)
from aniworld import parser


class CustomTheme(npyscreen.ThemeManager):
    default_colors = {
        'DEFAULT': 'WHITE_BLACK',
        'FORMDEFAULT': 'MAGENTA_BLACK',
        'NO_EDIT': 'BLUE_BLACK',
        'STANDOUT': 'CYAN_BLACK',
        'CURSOR': 'WHITE_BLACK',
        'CURSOR_INVERSE': 'BLACK_WHITE',
        'LABEL': 'CYAN_BLACK',
        'LABELBOLD': 'CYAN_BLACK',
        'CONTROL': 'GREEN_BLACK',
        'IMPORTANT': 'GREEN_BLACK',
        'SAFE': 'GREEN_BLACK',
        'WARNING': 'YELLOW_BLACK',
        'DANGER': 'RED_BLACK',
        'CRITICAL': 'BLACK_RED',
        'GOOD': 'GREEN_BLACK',
        'GOODHL': 'GREEN_BLACK',
        'VERYGOOD': 'BLACK_GREEN',
        'CAUTION': 'YELLOW_BLACK',
        'CAUTIONHL': 'BLACK_YELLOW',
    }


class SelectionMenu(npyscreen.NPSApp):
    def __init__(self, arguments, slug):
        super().__init__()
        self.arguments = arguments
        self.anime = Anime(slug=slug, episode_list=[
            Episode(slug=slug, season=1, episode=1)])
        self.selected_episodes = []
        self.episode_dict = {}
        self.action_selection = None
        self.aniskip_selection = None
        self.folder_selection = None
        self.language_selection = None
        self.provider_selection = None
        self.episode_selection = None
        self.select_all_button = None

    def main(self):  # pylint: disable=too-many-locals, too-many-statements
        available_languages = self.anime[0].language_name
        season_episode_count = self.anime[0].season_episode_count
        movie_episode_count = self.anime[0].movie_episode_count
        available_providers = self.anime[0].provider_name

        supported_providers = [
            provider for provider in available_providers if provider in SUPPORTED_PROVIDERS]

        for season, episodes in season_episode_count.items():
            for episode in range(1, episodes + 1):
                link_formatted = f"{self.anime.title} - Season {season} - Episode {episode}"
                link = (
                    f"{ANIWORLD_TO}/anime/stream/{self.anime.slug}/"
                    f"staffel-{season}/episode-{episode}"
                )
                self.episode_dict[link] = link_formatted

        for episode in range(1, movie_episode_count + 1):
            movie_link_formatted = f"{self.anime.title} - Movie {episode}"
            movie_link = f"{ANIWORLD_TO}/anime/stream/{self.anime.slug}/filme/film-{episode}"
            self.episode_dict[movie_link] = movie_link_formatted

        available_episodes = list(self.episode_dict.values())

        terminal_height = os.get_terminal_size().lines

        # these are the heights of each widget
        total_reserved_height = (
            3 + 2 + 2 + 2 + max(2, len(available_languages)) +
            max(2, len(supported_providers)) + 5
        )

        max_episode_height = max(3, terminal_height - total_reserved_height)

        npyscreen.setTheme(CustomTheme)
        form_title = (
            f"Welcome to AniWorld-Downloader v.{VERSION}"
            f"{' (Update Available)' if not IS_NEWEST_VERSION else ''}"
        )
        f = npyscreen.Form(name=form_title)

        self.action_selection = f.add(
            npyscreen.TitleSelectOne,
            max_height=3,
            value=[["Watch", "Download", "Syncplay"].index(
                self.arguments.action)],
            name="Action",
            values=["Watch", "Download", "Syncplay"],
            scroll_exit=True
        )

        self.aniskip_selection = f.add(
            npyscreen.TitleMultiSelect,
            max_height=2,
            name="Aniskip",
            values=["Enabled"],
            scroll_exit=True,
            rely=self.action_selection.rely + self.action_selection.height + 1
        )

        self.folder_selection = f.add(
            npyscreen.TitleFilenameCombo,
            max_height=2,
            name="Save Location",
            rely=self.action_selection.rely + self.action_selection.height + 1,
            value=self.arguments.output_dir
        )

        self.language_selection = f.add(
            npyscreen.TitleSelectOne,
            max_height=max(2, len(available_languages)),
            value=[
                available_languages.index(self.arguments.language)
                if self.arguments.language in available_languages else 0
            ],
            name="Language",
            values=available_languages,
            scroll_exit=True,
            rely=self.aniskip_selection.rely + self.aniskip_selection.height
        )

        self.provider_selection = f.add(
            npyscreen.TitleSelectOne,
            max_height=max(2, len(supported_providers)),
            value=[
                supported_providers.index(self.arguments.provider)
                if self.arguments.provider in supported_providers else 0
            ],
            name="Provider",
            values=supported_providers,
            scroll_exit=True,
            rely=self.language_selection.rely + self.language_selection.height + 1
        )

        self.episode_selection = f.add(
            npyscreen.TitleMultiSelect,
            max_height=max_episode_height,
            name="Episode",
            values=available_episodes,
            scroll_exit=True,
            rely=self.provider_selection.rely + self.provider_selection.height + 1
        )

        self.select_all_button = f.add(
            npyscreen.ButtonPress,
            name="Select All",
            rely=self.episode_selection.rely + self.episode_selection.height + 1
        )

        def toggle_select_all():
            if len(self.episode_selection.value) == len(available_episodes):
                self.episode_selection.value = []
                self.selected_episodes = []
                self.select_all_button.name = "Select All"
            else:
                self.episode_selection.value = list(
                    range(len(available_episodes)))
                self.selected_episodes = list(self.episode_dict.keys())
                self.select_all_button.name = "Deselect All"
            f.display()

        self.select_all_button.whenPressed = toggle_select_all

        def update_visibility():
            selected_action = self.action_selection.get_selected_objects()[0]
            if selected_action in ["Watch", "Syncplay"]:
                self.folder_selection.hidden = True
                self.aniskip_selection.hidden = False

                if USES_DEFAULT_PROVIDER:
                    try:
                        provider_index = supported_providers.index(
                            DEFAULT_PROVIDER_WATCH)
                    except ValueError:
                        provider_index = 0

                    if self.provider_selection.value != [provider_index]:
                        self.provider_selection.value = [provider_index]
            else:
                self.folder_selection.hidden = False
                self.aniskip_selection.hidden = True

                if USES_DEFAULT_PROVIDER:
                    try:
                        provider_index = supported_providers.index(
                            DEFAULT_PROVIDER_DOWNLOAD)
                    except ValueError:
                        provider_index = 0

                    if self.provider_selection.value != [provider_index]:
                        self.provider_selection.value = [provider_index]
            f.display()

        self.action_selection.when_value_edited = update_visibility

        update_visibility()

        self.episode_selection.when_value_edited = self.on_ok

        f.edit()

    def on_ok(self):
        selected_link_formatted = self.episode_selection.get_selected_objects() or []

        # self.selected_episodes = [
        #    {"link": link, "name": name}
        #    for link, name in self.episode_dict.items() if name in selected_link_formatted
        # ]

        self.selected_episodes = [
            link
            for link, name in self.episode_dict.items() if name in selected_link_formatted
        ]

    def get_selected_values(self):
        """
        return Anime(
            title=self.anime.title,
            episode_list=[
                Episode(
                    episode=episode["name"],
                    slug=self.anime.slug,
                    link=episode["link"],
                    # _selected_provider=self.anime.provider,
                    # _selected_language=self.anime.language
                    _selected_provider="VOE",
                    _selected_language="German Sub"
                ) for episode in self.selected_episodes
            ]
        )
        """

        # return self.selected_episodes

        # return Anime(
        #     action=self.action_selection.get_selected_objects()[0],
        #     language=self.language_selection.get_selected_objects()[0],
        #     provider=self.provider_selection.get_selected_objects()[0],
        #     output_directory=self.folder_selection.value,
        #     episode_list=self.selected_episodes
        # )

        selected_action = self.action_selection.get_selected_objects()[0]
        selected_language = self.language_selection.get_selected_objects()[0]
        selected_provider = self.provider_selection.get_selected_objects()[0]
        selected_output_directory = self.folder_selection.value
        selected_aniskip = bool(self.aniskip_selection.value)
        parser.arguments.action = selected_action

        # print(f"Anime Title: {self.anime.title}")
        # print(f"Selected Episodes: {self.selected_episodes}")
        # print(f"Action: {selected_action}")
        # print(f"Language: {selected_language}")
        # print(f"Provider: {selected_provider}")
        # print(f"Output Directory: {selected_output_directory}")

        log_message = (
            "Selected Values:\n"
            f"Selected Action: {selected_action}\n"
        )

        if selected_action == "Watch":
            log_message += f"Selected Aniskip: {selected_aniskip}\n"
        elif selected_action == "Download":
            log_message += f"Selected Output Directory: {selected_output_directory}\n"

        log_message += (
            f"Selected Language: {selected_language}\n"
            f"Selected Provider: {selected_provider}\n"
            f"Selected Episode: {self.selected_episodes}\n"
        )

        logging.debug(log_message)

        try:
            return Anime(
                title=self.anime.title,
                episode_list=[
                    Episode(
                        slug=self.anime.slug,
                        link=link,
                        _selected_language=selected_language,
                        _selected_provider=selected_provider
                    ) for link in self.selected_episodes
                ],
                action=selected_action,
                language=selected_language,
                provider=selected_provider,
                output_directory=selected_output_directory,
                aniskip=selected_aniskip
            )
        except urllib3.exceptions.ReadTimeoutError:
            print("Request timed out. Please try again later or use a VPN.")
            sys.exit()


def menu(arguments, slug):
    # print(arguments)
    try:
        app = SelectionMenu(arguments=arguments, slug=slug)
        app.run()
        anime = app.get_selected_values()
    except KeyboardInterrupt:
        curses.endwin()
        sys.exit()

    # curses.endwin()
    return anime


if __name__ == "__main__":
    selected_episodes = menu(slug="dan-da-dan", arguments=None)
    print("Selected Episodes:", selected_episodes)



import re

import requests
from bs4 import BeautifulSoup

from aniworld.config import DEFAULT_REQUEST_TIMEOUT, RANDOM_USER_AGENT


def get_direct_link_from_vidoza(embeded_vidoza_link: str) -> str:
    response = requests.get(
        embeded_vidoza_link,
        headers={'User-Agent': RANDOM_USER_AGENT},
        timeout=DEFAULT_REQUEST_TIMEOUT
    )

    soup = BeautifulSoup(response.content, "html.parser")

    for tag in soup.find_all('script'):
        if 'sourcesCode:' in tag.text:
            match = re.search(r'src: "(.*?)"', tag.text)
            if match:
                return match.group(1)

    raise ValueError("No direct link found.")


if __name__ == '__main__':
    link = input("Enter Vidoza Link: ")
    print(get_direct_link_from_vidoza(embeded_vidoza_link=link))



import re

import requests
from bs4 import BeautifulSoup

from aniworld.config import DEFAULT_REQUEST_TIMEOUT, RANDOM_USER_AGENT


def get_direct_link_from_vidmoly(embeded_vidmoly_link: str):
    response = requests.get(
        embeded_vidmoly_link,
        headers={'User-Agent': RANDOM_USER_AGENT},
        timeout=DEFAULT_REQUEST_TIMEOUT
    )
    html_content = response.text
    soup = BeautifulSoup(html_content, 'html.parser')
    scripts = soup.find_all('script')

    file_link_pattern = r'file:\s*"(https?://.*?)"'

    for script in scripts:
        if script.string:
            match = re.search(file_link_pattern, script.string)
            if match:
                file_link = match.group(1)
                return file_link

    raise ValueError("No direct link found.")


if __name__ == '__main__':
    link = input("Enter Vidmoly Link: ")
    print('Note: --referer "https://vidmoly.to"')
    print(get_direct_link_from_vidmoly(embeded_vidmoly_link=link))



from .provider.voe import get_direct_link_from_voe
from .provider.vidoza import get_direct_link_from_vidoza
from .provider.vidmoly import get_direct_link_from_vidmoly
from .provider.doodstream import get_direct_link_from_doodstream
from .provider.speedfiles import get_direct_link_from_speedfiles
from .provider.luluvdo import get_direct_link_from_luluvdo
from .provider.hanime import get_direct_link_from_hanime



import re

import requests

from aniworld import config


def get_direct_link_from_luluvdo(embeded_luluvdo_link, arguments=None):
    luluvdo_id = embeded_luluvdo_link.split('/')[-1]
    filelink = (
        f"https://luluvdo.com/dl?op=embed&file_code={luluvdo_id}&embed=1&referer=luluvdo.com&adb=0"
    )

    # The User-Agent needs to be the same as the direct-link ones to work
    headers = {
        "Origin": "https://luluvdo.com",
        "Referer": "https://luluvdo.com/",
        "User-Agent": config.LULUVDO_USER_AGENT
    }

    if arguments.action == "Download":
        headers["Accept-Language"] = "de-DE,de;q=0.9,en-US;q=0.8,en;q=0.7"

    response = requests.get(filelink, headers=headers,
                            timeout=config.DEFAULT_REQUEST_TIMEOUT)

    if response.status_code == 200:
        pattern = r'file:\s*"([^"]+)"'
        matches = re.findall(pattern, str(response.text))

        if matches:
            return matches[0]

    raise ValueError("No match found")


if __name__ == '__main__':
    url = input("Enter Luluvdo Link: ")
    print(get_direct_link_from_luluvdo(url))



import re
import json
import sys
import requests
from aniworld.config import DEFAULT_REQUEST_TIMEOUT


def fetch_page_content(url):
    try:
        response = requests.get(url, timeout=DEFAULT_REQUEST_TIMEOUT)
        response.raise_for_status()
        return response.text
    except requests.exceptions.RequestException as e:
        print(f"Failed to fetch the page content: {e}")
        return None


def extract_video_data(page_content):
    match = re.search(r'^.*videos_manifest.*$', page_content, re.MULTILINE)
    if not match:
        raise ValueError("Failed to extract video manifest from the response.")

    json_str = match.group(0)[match.group(0).find(
        '{'):match.group(0).rfind('}') + 1]
    return json.loads(json_str)


def get_streams(url):
    page_content = fetch_page_content(url)
    data = extract_video_data(page_content)
    video_info = data['state']['data']['video']
    name = video_info['hentai_video']['name']
    streams = video_info['videos_manifest']['servers'][0]['streams']

    return {"name": name, "streams": streams}


def display_streams(streams):
    if not streams:
        print("No streams available.")
        return

    print("Available qualities:")
    for i, stream in enumerate(streams, 1):
        premium_tag = "(Premium)" if not stream['is_guest_allowed'] else ""
        print(
            f"{i}. {stream['width']}x{stream['height']}\t"
            f"({stream['filesize_mbs']}MB) {premium_tag}")


def get_user_selection(streams):
    try:
        selected_index = int(input("Select a stream: ").strip()) - 1
        if 0 <= selected_index < len(streams):
            return selected_index

        print("Invalid selection.")
        return None
    except ValueError:
        print("Invalid input.")
        return None


def get_direct_link_from_hanime(url=None):
    try:
        if url is None:
            if len(sys.argv) > 1:
                url = sys.argv[1]
            else:
                url = input("Please enter the hanime.tv video URL: ").strip()

        try:
            video_data = get_streams(url)
            print(f"Video: {video_data['name']}")
            print('*' * 40)
            display_streams(video_data['streams'])

            selected_index = None
            while selected_index is None:
                selected_index = get_user_selection(video_data['streams'])

            print(f"M3U8 URL: {video_data['streams'][selected_index]['url']}")
        except ValueError as e:
            print(f"Error: {e}")
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    get_direct_link_from_hanime()



import traceback
import logging
import sys

from aniworld.ascii_art import display_traceback_art
from aniworld.action import watch, syncplay
from aniworld.models import Anime, Episode
from aniworld.parser import arguments
from aniworld.search import search_anime
from aniworld.execute import execute
from aniworld.menu import menu
from aniworld.common import generate_links


def aniworld() -> None:  # pylint: disable=too-many-branches, too-many-statements
    try:
        if arguments.local_episodes:
            if arguments.action == "Watch":
                watch(None)
            elif arguments.action == "Syncplay":
                syncplay(None)
        if arguments.episode or arguments.episode_file:
            links = []
            if arguments.episode_file:
                try:
                    with open(arguments.episode_file, 'r', encoding="UTF-8") as file:
                        urls = []
                        for line in file:
                            line = line.strip()
                            if line.startswith("http"):
                                urls.append(line)

                        links.extend(urls)
                except FileNotFoundError:
                    logging.error(
                        "The specified episode file does not exist: %s", arguments.episode_file
                    )
                    sys.exit(1)
                except IOError as e:
                    logging.error("Error reading the episode file: %s", e)
                    sys.exit(1)

            if arguments.episode:
                links.extend(arguments.episode)
            links = generate_links(links, arguments)

            # TODO: this needs to pass all links to a function
            #       that will return Anime objects instead
            anime_list = []
            for link in (links or [None]):
                if link:
                    episode = Episode(
                        link=link
                    )
                else:
                    slug = arguments.slug if arguments.slug else search_anime()
                    episode = Episode(
                        slug=slug
                    )
                anime = Anime(episode_list=[episode])
                anime_list.append(anime)

            execute(anime_list=anime_list)
        if not arguments.episode and not arguments.local_episodes and not arguments.episode_file:
            while True:
                try:
                    slug = arguments.slug if arguments.slug else search_anime()
                    break
                except ValueError:
                    continue

            anime = menu(arguments=arguments, slug=slug)
            execute(anime_list=[anime])
    except KeyboardInterrupt:
        pass
    except Exception as e:  # pylint: disable=broad-exception-caught
        if arguments.debug:
            traceback.print_exc()
        else:
            # hide traceback only show output
            print(display_traceback_art())
            print(f"Error: {e}")
            print("\nFor more detailed information, use --debug and try again.")

        # Detecting Nuitka at run time
        if "__compiled__" in globals():
            input("Press Enter to exit...")


if __name__ == "__main__":
    aniworld()



import re
import base64
import json
from urllib.request import Request, urlopen
from urllib.error import HTTPError, URLError

import requests
from bs4 import BeautifulSoup

from aniworld import config


def shift_letters(input_str):
    result = ''
    for c in input_str:
        code = ord(c)
        if 65 <= code <= 90:
            code = (code - 65 + 13) % 26 + 65
        elif 97 <= code <= 122:
            code = (code - 97 + 13) % 26 + 97
        result += chr(code)
    return result


def replace_junk(input_str):
    junk_parts = ['@$', '^^', '~@', '%?', '*~', '!!', '#&']
    for part in junk_parts:
        input_str = re.sub(re.escape(part), '_', input_str)
    return input_str


def shift_back(s, n):
    return ''.join(chr(ord(c) - n) for c in s)


def decode_voe_string(encoded):
    step1 = shift_letters(encoded)
    step2 = replace_junk(step1).replace('_', '')
    step3 = base64.b64decode(step2).decode()
    step4 = shift_back(step3, 3)
    step5 = base64.b64decode(step4[::-1]).decode()
    return json.loads(step5)


def extract_voe_from_script(html):
    soup = BeautifulSoup(html, "html.parser")
    script = soup.find("script", type="application/json")
    return decode_voe_string(script.text[2:-2])["source"]


def get_direct_link_from_voe(embeded_voe_link: str) -> str:
    response = requests.get(
        embeded_voe_link,
        headers={'User-Agent': config.RANDOM_USER_AGENT},
        timeout=config.DEFAULT_REQUEST_TIMEOUT
    )

    redirect = re.search(r"https?://[^'\"<>]+", response.text)
    if not redirect:
        raise ValueError("No redirect found.")

    redirect_url = redirect.group(0)
    parts = redirect_url.strip().split("/")
    config.PROVIDER_HEADERS["VOE"].append(
        f'Referer: "{parts[0]}//{parts[2]}/"')

    try:
        with urlopen(
            Request(
                redirect_url,
                headers={'User-Agent': config.RANDOM_USER_AGENT}
            ),
            timeout=config.DEFAULT_REQUEST_TIMEOUT
        ) as resp:
            html = resp.read().decode()
    except (HTTPError, URLError, TimeoutError) as err:
        raise ValueError(f"Redirect failed: {err}") from err

    extracted = extract_voe_from_script(html)
    if extracted:
        return extracted

    b64match = re.search(r"var a168c='([^']+)'", html)
    if b64match:
        decoded = base64.b64decode(b64match.group(1)).decode()[::-1]
        return json.loads(decoded)["source"]

    hls = re.search(r"'hls': '(?P<hls>[^']+)'", html)
    if hls:
        return base64.b64decode(hls.group("hls")).decode()

    raise ValueError(
        "Unable to process this VOE link.\n\n"
        "Try using a different provider for now.\n"
        "If this issue persists and hasn't been reported yet, please consider creating a new issue."
    )


if __name__ == '__main__':
    link = input("Enter VOE Link: ")
    print(get_direct_link_from_voe(link))



import os
import platform
import logging
import subprocess
import shutil

from aniworld.config import MPV_DIRECTORY
from aniworld.common import get_github_release, download_file, remove_anime4k


def get_anime4k_download_link(mode: str = "High") -> str:
    os_type = "Windows" if platform.system() == "Windows" else "Mac_Linux"

    latest_release = get_github_release("Tama47/Anime4K")
    download_path = os.path.dirname(list(latest_release.values())[0])
    download_link = f"{download_path}/GLSL_{os_type}_{mode}-end.zip"

    return download_link


def download_anime4k(mode):
    if mode == "Remove":
        remove_anime4k()
        return

    os.makedirs(MPV_DIRECTORY, exist_ok=True)
    archive_path = os.path.join(MPV_DIRECTORY, "anime4k.zip")

    if not os.path.exists(archive_path):
        download_link = get_anime4k_download_link(mode)
        print("Downloading Anime4K...")
        download_file(download_link, archive_path)

        extract_anime4k(zip_path=archive_path, dep_path=MPV_DIRECTORY)
    else:
        logging.warning("File already exists at: %s", archive_path)


def extract_anime4k(zip_path, dep_path):
    logging.debug("Unpacking Anime4K to %s", dep_path)
    try:
        subprocess.run(
            ["tar", "-xf", zip_path],
            check=True,
            cwd=dep_path
        )
    except subprocess.CalledProcessError as e:
        logging.error("Failed to extract files: %s", e)
    except subprocess.SubprocessError as e:
        logging.error("An error occurred: %s", e)

    if os.path.exists(zip_path):
        os.remove(zip_path)

    anime4k_path = os.path.join(MPV_DIRECTORY, "__MACOSX")
    if os.path.exists(anime4k_path):
        shutil.rmtree(anime4k_path)


if __name__ == "__main__":
    download_anime4k("High")



import os
import re
import subprocess
import logging

from aniworld.models import Anime
from aniworld.config import PROVIDER_HEADERS, INVALID_PATH_CHARS
from aniworld.parser import arguments


def download(anime: Anime):  # pylint: disable=too-many-branches
    for episode in anime:
        episode_title = f"{anime.title} - S{episode.season}E{episode.episode} - ({anime.language}):"
        direct_link = episode.get_direct_link()

        if not direct_link:
            logging.warning(f"Something went wrong with \"{episode_title}\".")
            continue

        if arguments.only_direct_link:
            print(episode_title)
            print(f"{direct_link}\n")
            continue

        sanitized_anime_title = ''.join(
            char for char in anime.title if char not in INVALID_PATH_CHARS
        )

        if episode.season == 0:
            output_file = (
                f"{sanitized_anime_title} - "
                f"Movie {episode.episode:02} - "
                f"({anime.language}).mp4"
            )
        else:
            output_file = (
                f"{sanitized_anime_title} - "
                f"S{episode.season:02}E{episode.episode:03} - "
                f"({anime.language}).mp4"
            )

        output_path = os.path.join(
            arguments.output_dir, sanitized_anime_title, output_file
        )

        os.makedirs(os.path.dirname(output_path), exist_ok=True)

        command = [
            "yt-dlp",
            direct_link,
            "--fragment-retries", "infinite",
            "--concurrent-fragments", "4",
            "-o", output_path,
            "--quiet",
            "--no-warnings",
            "--progress"
        ]

        if anime.provider in PROVIDER_HEADERS:
            for header in PROVIDER_HEADERS[anime.provider]:
                command.extend(["--add-header", header])

        if arguments.only_command:
            print(
                f"\n{anime.title} - S{episode.season}E{episode.episode} - ({anime.language}):"
            )
            print(
                f"{' '.join(str(item) if item is not None else '' for item in command)}"
            )
            continue

        try:
            print(f"Downloading to {output_path}...")
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError:
            print(
                "Error running command:\n"
                f"{' '.join(str(item) if item is not None else '' for item in command)}"
            )
        except KeyboardInterrupt:
            # directory containing the output_path
            output_dir = os.path.dirname(output_path)
            is_empty = True

            # delete all .part, .ytdl, or .part-Frag followed by any number in output_path
            for file_name in os.listdir(output_dir):
                if re.search(r'\.(part|ytdl|part-Frag\d+)$', file_name):
                    os.remove(os.path.join(output_dir, file_name))
                else:
                    is_empty = False

            # delete folder too if empty after
            if is_empty or not os.listdir(output_dir):
                os.rmdir(output_dir)



import re
import base64
import requests

from aniworld.config import DEFAULT_REQUEST_TIMEOUT, RANDOM_USER_AGENT

SPEEDFILES_PATTERN = re.compile(r'var _0x5opu234 = "(?P<encoded_data>.*?)";')


def get_direct_link_from_speedfiles(embeded_speedfiles_link):
    response = requests.get(
        embeded_speedfiles_link,
        timeout=DEFAULT_REQUEST_TIMEOUT,
        headers={'User-Agent': RANDOM_USER_AGENT}
    )

    if "<span class=\"inline-block\">Web server is down</span>" in response.text:
        raise ValueError(
            "The SpeedFiles server is currently down.\n"
            "Please try again later or choose a different hoster."
        )

    match = SPEEDFILES_PATTERN.search(response.text)

    if not match:
        raise ValueError("Pattern not found in the response.")

    encoded_data = match.group("encoded_data")
    decoded = base64.b64decode(encoded_data).decode()
    decoded = decoded.swapcase()[::-1]
    decoded = base64.b64decode(decoded).decode()[::-1]
    decoded_hex = ''.join(chr(int(decoded[i:i + 2], 16))
                          for i in range(0, len(decoded), 2))
    shifted = ''.join(chr(ord(char) - 3) for char in decoded_hex)
    result = base64.b64decode(shifted.swapcase()[::-1]).decode()

    return result


if __name__ == '__main__':
    speedfiles_link = input("Enter Speedfiles Link: ")
    print(get_direct_link_from_speedfiles(
        embeded_speedfiles_link=speedfiles_link))



import requests
from bs4 import BeautifulSoup

from aniworld.config import DEFAULT_REQUEST_TIMEOUT, RANDOM_USER_AGENT

# Currently not working cause of capture


def get_direct_link_from_filemoon(embeded_filemoon_link: str) -> str:
    filemoon_id = embeded_filemoon_link.split('/')[-1]
    filemoon_link = f"https://filemoon.to/download/{filemoon_id}"
    response = requests.get(
        filemoon_link,
        headers={'User-Agent': RANDOM_USER_AGENT},
        timeout=DEFAULT_REQUEST_TIMEOUT
    )
    soup = BeautifulSoup(response.content, 'html.parser')

    download_link = soup.find('a', class_='button')['href']
    return download_link


if __name__ == '__main__':
    link = input("Enter Filemoon Link: ")
    print(get_direct_link_from_filemoon(embeded_filemoon_link=link))



import re
import json
import logging

import requests
import requests.models
from bs4 import BeautifulSoup

from aniworld.aniskip import get_mal_id_from_title
from aniworld.config import DEFAULT_REQUEST_TIMEOUT, RANDOM_USER_AGENT, ANIWORLD_TO
from aniworld.parser import arguments
from aniworld.common import get_season_episode_count, get_movie_episode_count

from aniworld.extractors import (
    get_direct_link_from_vidmoly,
    get_direct_link_from_vidoza,
    get_direct_link_from_voe,
    get_direct_link_from_doodstream,
    get_direct_link_from_speedfiles,
    get_direct_link_from_luluvdo
)


# Implement lazy loading for the Anime and Episode class.
# Initialize variables with None.
# When a variable is accessed, fetch and store its value automatically,
# so it doesn't need to be fetched again on subsequent accesses.


class Anime:
    """
    Represents an anime series with various attributes and methods to fetch and manage its details.

    Example:
        anime = Anime(
            episode_list=[
                Episode(
                    slug="loner-life-in-another-world",
                    season=1,
                    episode=1
                )
            ]
        )

    Required Attributes:
        episode_list (list): A list of Episode objects for the anime.

    Attributes:
        title (str): The title of the anime.
        slug (str): A URL-friendly version of the title used for web requests.
        action (str): The default action to be performed.
                      Must be one of "Download", "Watch", or "Syncplay". 
                      Defaults to "Watch" if not provided.
        provider (str): The provider of the anime content. Defaults to:
                        - DEFAULT_PROVIDER_DOWNLOAD if action is "Download"
                        - DEFAULT_PROVIDER_WATCH if action is not "Download"
                        Set based on the action if not provided.
        language (str): The language code for the anime.
                        Defaults to DEFAULT_LANGUAGE if not provided.
        aniskip (bool): Whether to skip certain actions (default is False).
        output_directory (str): The directory where downloads are saved.
                                Defaults to the user's home "Downloads" directory.
        episode_list (list): A list of Episode objects for the anime.
        description_german (str): The German description of the anime.
        description_english (str): The English description of the anime.
        html (requests.models.Response): The HTML response object for the anime's webpage.
    """

    def __init__(
        self,
        title=None,
        slug=None,
        action=arguments.action,
        provider=arguments.provider,
        language=arguments.language,
        aniskip=arguments.aniskip,
        output_directory=arguments.output_dir,
        episode_list=None,
        description_german=None,
        # description_english=None,
        html=None
    ) -> None:
        if not episode_list:
            raise ValueError("Provide 'episode_list'.")

        self.slug = slug or episode_list[0].slug
        if not self.slug:
            raise ValueError("Slug of Anime is None.")

        self.html = html or requests.get(
            f"{ANIWORLD_TO}/anime/stream/{self.slug}",
            timeout=DEFAULT_REQUEST_TIMEOUT
        )

        self.title = title or get_anime_title_from_html(self.html)
        self.action = action
        self.provider = provider
        self.language = language
        self.aniskip = aniskip
        self.output_directory = output_directory
        self.episode_list = episode_list

        self.description_german = description_german or self._fetch_description_german()
        # This should be fetched manually as its not needed for now
        # self.description_english = description_english or self._fetch_description_english()

    def _fetch_description_german(self):
        soup = BeautifulSoup(self.html.content, 'html.parser')
        desc_div = soup.find('p', class_='seri_des')
        return (
            desc_div.get('data-full-description', '')
            if desc_div else "Could not fetch description."
        )

    def _fetch_description_english(self):
        anime_id = get_mal_id_from_title(self.title, 1)
        response = requests.get(
            f"https://myanimelist.net/anime/{anime_id}",
            timeout=DEFAULT_REQUEST_TIMEOUT
        )
        soup = BeautifulSoup(response.content, 'html.parser')
        desc_meta = soup.find('meta', property='og:description')
        return (
            desc_meta['content']
            if desc_meta else "Could not fetch description."
        )

    def __iter__(self):
        return iter(self.episode_list)

    def __getitem__(self, index: int):
        return self.episode_list[index]

    def to_json(self) -> str:
        data = {
            "title": self.title,
            "slug": self.slug,
            "action": self.action,
            "provider": self.provider,
            "language": self.language,
            "aniskip": self.aniskip,
            "output_directory": str(self.output_directory),
            "episode_list": self.episode_list,
            "description_german": ' '.join(self.description_german.split()[:10]) + ' [...]',
            # "description_english": ' '.join(self.description_english.split()[:10]) + ' [...]',
        }
        return str(data)

    def __str__(self):
        return self.to_json()


class Episode:
    """
    Represents an episode of an anime series with various attributes
    and methods to fetch and manage its details.

    Example:
        Episode(
            slug="loner-life-in-another-world",
            season=1,
            episode=1
        )

    Required Attributes:
        link (str) or slug (str), season (int), episode (int):
        Either a direct link to the episode or a slug with season
        and episode numbers for constructing the link.

    Attributes:
        anime_title (str): The title of the anime the episode belongs to.
        title_german (str): The German title of the episode.
        title_english (str): The English title of the episode.
        season (int): The season number of the episode.
        episode (int): The episode number within the season.
        slug (str): A URL-friendly version of the episode title used for web requests.
        link (str): The direct link to the episode.
        mal_id (int): The MyAnimeList ID for the episode.
        redirect_link (str): The redirect link for streaming the episode.
        embeded_link (str): The embedded link for the episode.
        direct_link (str): The direct streaming link for the episode.
        provider (dict): A dictionary of available providers and their links.
        provider_name (list): A list of provider names.
        language (list): A list of available language codes for the episode.
        language_name (list): A list of available language names for the episode.
        season_episode_count (dict): A dictionary mapping season numbers to episode counts.
        movie_episode_count (int): The count of movie episodes.
        html (requests.models.Response): The HTML response object for the episode's webpage.
        _selected_provider (str): The selected provider for streaming.
        _selected_language (int): The selected language code for streaming.
    """

    def __init__(  # pylint: disable=too-many-locals
        self,
        anime_title: str = None,
        title_german: str = None,
        title_english: str = None,
        season: int = None,
        episode: int = None,
        slug: str = None,
        link: str = None,
        mal_id: int = None,
        redirect_link: str = None,
        embeded_link: str = None,
        direct_link: str = None,
        provider: dict = None,  # available providers
        provider_name: list = None,
        language: list = None,  # available languages
        language_name: list = None,
        season_episode_count: dict = None,
        has_movies: bool = False,
        movie_episode_count: int = None,
        html: requests.models.Response = None,
        _selected_provider: str = arguments.provider,
        _selected_language: str = arguments.language
    ) -> None:
        if not link and (not slug or not season or not episode):
            raise ValueError(
                "Provide either 'link' or 'slug' with 'season' and 'episode'.")

        self.anime_title: str = anime_title
        self.title_german: str = title_german
        self.title_english: str = title_english
        self.season: int = season
        self.episode: int = episode
        self.slug: str = slug
        self.link: str = link
        self.mal_id: int = mal_id
        self.redirect_link = redirect_link
        self.embeded_link = embeded_link
        self.direct_link = direct_link
        self.provider: dict = provider
        self.provider_name: list = provider_name
        self.language: list = language
        self.language_name: list = language_name
        self.season_episode_count: dict = season_episode_count
        self.has_movies: bool = has_movies
        self.movie_episode_count: int = movie_episode_count
        self.html: requests.models.Response = html
        self._selected_provider: str = _selected_provider
        self._selected_language: str = _selected_language

        self.auto_fill_details()

    def _get_episode_title_from_html(self) -> tuple:
        episode_soup = BeautifulSoup(self.html.content, 'html.parser')
        episode_german_title_div = episode_soup.find(
            'span', class_='episodeGermanTitle')
        episode_english_title_div = episode_soup.find(
            'small', class_='episodeEnglishTitle')

        german_title = episode_german_title_div.text if episode_german_title_div else ""
        english_title = episode_english_title_div.text if episode_english_title_div else ""

        return german_title, english_title

    def _get_season_from_link(self) -> int:
        if "/filme/" in self.link:
            return 0  # Set season to 0 for movies

        season = self.link.split("/")[-2]  # e.g. staffel-2
        numbers = re.findall(r'\d+', season)

        if numbers:
            return int(numbers[-1])  # e.g 2

        raise ValueError(
            f"No valid season number found in the link: {self.link}")

    def _get_episode_from_link(self) -> int:
        if self.link.endswith("/"):
            self.link = self.link[:-1]

        episode = self.link.split("/")[-1]  # e.g. episode-2
        numbers = re.findall(r'\d+', episode)

        if numbers:
            return int(numbers[-1])  # e.g 2

        raise ValueError(
            f"No valid episode number found in the link: {self.link}")

    def _get_available_language_from_html(self) -> list[int]:
        """
        Language Codes:
            1: German Dub
            2: English Sub
            3: German Sub
        """

        episode_soup = BeautifulSoup(self.html.content, 'html.parser')
        change_language_box_div = episode_soup.find(
            'div', class_='changeLanguageBox')
        language = []

        if change_language_box_div:
            img_tags = change_language_box_div.find_all('img')
            for img in img_tags:
                lang_key = img.get('data-lang-key')
                if lang_key and lang_key.isdigit():
                    language.append(int(lang_key))
        else:
            return []

        return language  # e.g. [1, 2, 3]

    def _get_provider_from_html(self) -> dict:
        """
            Parses the HTML content to extract streaming providers,
            their language keys, and redirect links.

            Returns a dictionary with provider names as keys
            and language key-to-redirect URL mappings as values.

            Example:

            {
                'VOE': {1: 'https://aniworld.to/redirect/1766412',
                        2: 'https://aniworld.to/redirect/1766405'},
                'Doodstream': {1: 'https://aniworld.to/redirect/1987922',
                               2: 'https://aniworld.to/redirect/2700342'},
                ...
            }

            Access redirect link with:
            print(self.provider["VOE"][2])
        """

        soup = BeautifulSoup(self.html.content, 'html.parser')
        providers = {}

        episode_links = soup.find_all(
            'li', class_=lambda x: x and x.startswith('episodeLink')
        )

        if not episode_links:
            raise ValueError(
                f"{self.link}\nDerzeit keine Streams für diese Episode verfügbar. "
                f"Versuche es später noch einmal oder frage in der Shoutbox etc. nach ;)"
            )

        for link in episode_links:
            provider_name_tag = link.find('h4')
            provider_name = provider_name_tag.text.strip() if provider_name_tag else None

            redirect_link_tag = link.find('a', class_='watchEpisode')
            redirect_link = redirect_link_tag['href'] if redirect_link_tag else None

            lang_key = link.get('data-lang-key')
            lang_key = int(
                lang_key) if lang_key and lang_key.isdigit() else None

            if provider_name and redirect_link and lang_key:
                if provider_name not in providers:
                    providers[provider_name] = {}
                providers[provider_name][lang_key] = f"{ANIWORLD_TO}{redirect_link}"

        if not providers:
            raise ValueError(
                f"Could not get providers from {self.link}"
            )

        logging.debug("Available Episodes for \"%s\":\n%s",
                      self.anime_title, json.dumps(providers, indent=4))
        return providers

    def _get_key_from_language(self, language: str) -> int:
        lang_mapping = {
            "German Dub": 1,
            "English Sub": 2,
            "German Sub": 3
        }

        language_key = lang_mapping.get(language, None)

        if language_key is None:
            raise ValueError(f"Language: {language} not valid.")

        return language_key

    def _get_languages_from_keys(self, keys: list[int]) -> list[str]:
        key_mapping = {
            1: "German Dub",
            2: "English Sub",
            3: "German Sub"
        }

        languages = [key_mapping.get(key, None) for key in keys]

        if None in languages:
            raise ValueError("One or more keys are not valid.")

        return languages

    def _get_direct_link_from_provider(self) -> str:
        """
            providers = {
                "Vidmoly": get_direct_link_from_vidmoly,
                "Vidoza": get_direct_link_from_vidoza,
                "VOE": get_direct_link_from_voe,
                "Doodstream": get_direct_link_from_doodstream,
                "SpeedFiles": get_direct_link_from_speedfiles,
                "Luluvdo": get_direct_link_from_luluvdo
            }

            # first, try the selected provider
            if self._selected_provider in providers:
                try:
                    print(f"Trying selected provider: {self._selected_provider}")
                    return providers[self._selected_provider](self.embeded_link).get_direct_link()
                except ValueError as e:
                    print(f"An error occurred with {self._selected_provider}: {e}")

            # if the selected provider fails, try all other providers
            for provider_name, provider_func in providers.items():
                if provider_name == self._selected_provider:
                    continue  # skip the selected provider since we've already tried it

                try:
                    print(f"Trying {provider_name}")

                    # this fails because self.embeded_link is from the selected_provider
                    # and needs to be set to the next trying provider
                    direct_link = provider_func(self.embeded_link)

                    if direct_link is None:
                        raise ValueError

                except ValueError as e:
                    print(f"An error occurred with {provider_name}: {e}")
                    continue  # try the next provider if this one fails

            # if all providers fail, raise an error
            raise ValueError("All providers have failed.")
        """

        if self._selected_provider == "Vidmoly":
            return get_direct_link_from_vidmoly(embeded_vidmoly_link=self.embeded_link)
        if self._selected_provider == "Vidoza":
            return get_direct_link_from_vidoza(embeded_vidoza_link=self.embeded_link)
        if self._selected_provider == "VOE":
            return get_direct_link_from_voe(embeded_voe_link=self.embeded_link)
        if self._selected_provider == "Doodstream":
            return get_direct_link_from_doodstream(embeded_doodstream_link=self.embeded_link)
        if self._selected_provider == "SpeedFiles":
            return get_direct_link_from_speedfiles(embeded_speedfiles_link=self.embeded_link)
        if self._selected_provider == "Luluvdo":
            return get_direct_link_from_luluvdo(embeded_luluvdo_link=self.embeded_link,
                                                arguments=arguments)
        raise ValueError(
            f"{self._selected_provider} is currently not supported.")

    def get_redirect_link(self):
        lang_key = self._get_key_from_language(self._selected_language)

        if (self._selected_provider not in self.provider or
                lang_key not in self.provider[self._selected_provider]
            ):
            for provider_name, lang_dict in self.provider.items():
                if lang_key in lang_dict:
                    self._selected_provider = provider_name
                    self.redirect_link = lang_dict[lang_key]
                    break
            else:
                raise KeyError(
                    "No provider with the language key '"
                    f"{lang_key}' found. Checked providers: {list(self.provider.keys())}. "
                    f"Provider variable: {self.provider}"
                )
        else:
            self.redirect_link = self.provider[self._selected_provider][lang_key]

    def get_embeded_link(self):
        if not self.redirect_link:
            self.get_redirect_link()

        self.embeded_link = requests.get(
            self.redirect_link, timeout=DEFAULT_REQUEST_TIMEOUT,
            headers={'User-Agent': RANDOM_USER_AGENT}).url
        return self.embeded_link

    def get_direct_link(self, provider=None, language=None):
        """
        Retrieves the direct streaming link for the episode.

        Example:
            episode.get_direct_link("VOE", "German Sub")

        Note:
            If this method is not being called from a menu or with arguments,
            ensure that 'provider' and 'language' are set before calling.

        Args:
            provider (str): The name of the provider to use for fetching the direct link.
            language (str): The language code to use for fetching the direct link.

        Returns:
            str: The direct streaming link for the episode.
        """
        if provider:
            self._selected_provider = provider

        if language:
            self._selected_language = language

        if not self.embeded_link:
            self.get_embeded_link()

        self.direct_link = self._get_direct_link_from_provider()
        return self.direct_link

    def auto_fill_details(self) -> None:
        if self.slug and self.season and self.episode:
            self.link = (
                f"{ANIWORLD_TO}/anime/stream/{self.slug}/"
                f"staffel-{self.season}/episode-{self.episode}"
            )

        if self.link:
            self.slug = self.slug or self.link.split("/")[-3]
            self.season = self.season or self._get_season_from_link()
            self.episode = self.episode or self._get_episode_from_link()

        self.html = requests.get(self.link, timeout=DEFAULT_REQUEST_TIMEOUT)
        self.anime_title = get_anime_title_from_html(html=self.html)
        self.title_german, self.title_english = self._get_episode_title_from_html()
        self.language = self._get_available_language_from_html()
        self.language_name = self._get_languages_from_keys(self.language)
        self.provider = self._get_provider_from_html()
        self.provider_name = list(self.provider.keys())
        self.season_episode_count = get_season_episode_count(self.slug)
        self.movie_episode_count = get_movie_episode_count(self.slug)

        if self.movie_episode_count:
            # remove last season as its the same as movies and 0
            last_season = list(self.season_episode_count.keys())[-1]
            if self.season_episode_count[last_season] == 0:
                del self.season_episode_count[last_season]

    def to_json(self) -> str:
        data = {
            "anime_title": self.anime_title,
            "title_german": self.title_german,
            "title_english": self.title_english,
            "season": self.season,
            "episode": self.episode,
            "slug": self.slug,
            "link": self.link,
            "mal_id": self.mal_id,
            "redirect_link": self.redirect_link,
            "embeded_link": self.embeded_link,
            "direct_link": self.direct_link,
            "provider": self.provider,
            "provider_name": self.provider_name,
            "language": self.language,
            "language_name": self.language_name,
            "season_episode_count": self.season_episode_count,
            "movie_episode_count": self.movie_episode_count,
            "html": str(self.html)
        }
        return json.dumps(data, indent=4)

    def __str__(self) -> str:
        return self.to_json()


def get_anime_title_from_html(html: requests.models.Response):
    soup = BeautifulSoup(html.content, 'html.parser')
    title_div = soup.find('div', class_='series-title')

    if title_div:
        return title_div.find('h1').find('span').text

    return ""


if __name__ == "__main__":
    pass



from .common import (
    get_github_release,
    download_mpv,
    download_syncplay,
    download_file,
    get_season_episode_count,
    get_movie_episode_count,
    generate_links,
    remove_anime4k,
    remove_mpv_scripts,
    copy_file_if_different,
    setup_autostart,
    setup_autoexit
)



import logging
import os
import pathlib
import platform
import shutil
import tempfile
from importlib.metadata import PackageNotFoundError, version
from packaging.version import Version, InvalidVersion
from urllib3.exceptions import InsecureRequestWarning
import urllib3
import requests
from fake_useragent import UserAgent


#########################################################################################
# Global Constants
#########################################################################################

ANIWORLD_TO = "https://aniworld.to"
S_TO = "https://s.to"

#########################################################################################
# Logging Configuration
#########################################################################################

log_file_path = os.path.join(tempfile.gettempdir(), 'aniworld.log')


class CriticalErrorHandler(logging.Handler):
    def emit(self, record):
        if record.levelno == logging.CRITICAL:
            raise SystemExit(record.getMessage())


logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s:%(name)s:%(funcName)s: %(message)s",
    handlers=[
        logging.FileHandler(log_file_path, mode='w'),
        CriticalErrorHandler()
    ]
)

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.WARNING)
console_handler.setFormatter(logging.Formatter(
    "%(levelname)s:%(name)s:%(funcName)s: %(message)s")
)
logging.getLogger().addHandler(console_handler)
logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
logging.getLogger('charset_normalizer').setLevel(logging.WARNING)
logging.getLogger().setLevel(logging.WARNING)

urllib3.disable_warnings(InsecureRequestWarning)

#########################################################################################
# Default Configuration Constants
#########################################################################################

DEFAULT_REQUEST_TIMEOUT = 30

try:
    VERSION = version('aniworld')
except PackageNotFoundError:
    VERSION = ""


def get_latest_github_version():
    try:
        url = "https://api.github.com/repos/phoenixthrush/AniWorld-Downloader/releases/latest"
        response = requests.get(url, timeout=DEFAULT_REQUEST_TIMEOUT)
        return response.json().get('tag_name', '') if response.status_code == 200 else ""
    except requests.RequestException as e:
        logging.error("Error fetching latest release: %s", e)
        return ""


def is_newest_version():
    try:
        current = Version(VERSION.lstrip('v').lstrip('.'))
        latest_str = get_latest_github_version().lstrip('v').lstrip('.')
        latest = Version(latest_str)
        return latest, current >= latest
    except InvalidVersion as e:
        logging.error("Invalid version format: %s", e)
    except requests.RequestException as e:
        logging.error("Network error while fetching latest version: %s", e)

    return False


try:
    LATEST_VERSION, IS_NEWEST_VERSION = is_newest_version()
except TypeError:  # GitHub API Rate Limit (60/h) #52
    LATEST_VERSION = VERSION
    IS_NEWEST_VERSION = True

PLATFORM_SYSTEM = platform.system()

SUPPORTED_PROVIDERS = [
    "Doodstream", "Vidmoly", "Vidoza", "SpeedFiles", "Streamtape", "Luluvdo", "VOE"
]

#########################################################################################
# User Agents
#########################################################################################

RANDOM_USER_AGENT = UserAgent().random

LULUVDO_USER_AGENT = "Mozilla/5.0 (Android 15; Mobile; rv:132.0) Gecko/132.0 Firefox/132.0"

PROVIDER_HEADERS = {
    "Vidmoly": ['Referer: "https://vidmoly.to"'],
    "Doodstream": ['Referer: "https://dood.li/"'],
    "VOE": [f'User-Agent: {RANDOM_USER_AGENT}'],
    "Luluvdo": [
        f'User-Agent: {LULUVDO_USER_AGENT}',
        'Accept-Language: de-DE,de;q=0.9,en-US;q=0.8,en;q=0.7',
        'Origin: "https://luluvdo.com"',
        'Referer: "https://luluvdo.com/"'
    ]}

USES_DEFAULT_PROVIDER = False

# E.g. Watch, Download, Syncplay
DEFAULT_ACTION = "Download"
DEFAULT_ANISKIP = False
DEFAULT_DOWNLOAD_PATH = pathlib.Path.home() / "Downloads"
DEFAULT_KEEP_WATCHING = False
# German Dub, English Sub, German Sub
DEFAULT_LANGUAGE = "German Sub"
DEFAULT_ONLY_COMMAND = False
DEFAULT_ONLY_DIRECT_LINK = False
# SUPPORTED_PROVIDERS above
DEFAULT_PROVIDER_DOWNLOAD = "VOE"
DEFAULT_PROVIDER_WATCH = "SpeedFiles"
DEFAULT_TERMINAL_SIZE = (90, 30)

# https://learn.microsoft.com/en-us/windows/win32/fileio/naming-a-file
INVALID_PATH_CHARS = ['<', '>', ':', '"', '/', '\\', '|', '?', '*', '&']

#########################################################################################
# Executable Path Resolution
#########################################################################################

DEFAULT_APPDATA_PATH = os.path.join(
    os.getenv("APPDATA") or os.path.expanduser("~"), "aniworld"
)

if os.name == 'nt':
    MPV_DIRECTORY = os.path.join(os.environ.get('APPDATA', ''), 'mpv')
else:
    MPV_DIRECTORY = os.path.expanduser('~/.config/mpv')

MPV_SCRIPTS_DIRECTORY = os.path.join(MPV_DIRECTORY, 'scripts')

if platform.system() == "Windows":
    mpv_path = shutil.which("mpv")
    if not mpv_path:
        mpv_path = os.path.join(os.getenv('APPDATA', ''),
                                "aniworld", "mpv", "mpv.exe")
else:
    mpv_path = shutil.which("mpv")

MPV_PATH = mpv_path

if platform.system() == "Windows":
    syncplay_path = shutil.which("syncplay")
    if syncplay_path:
        syncplay_path = syncplay_path.replace(
            "syncplay.EXE", "SyncplayConsole.exe")
    else:
        syncplay_path = os.path.join(
            os.getenv(
                'APPDATA', ''), "aniworld", "syncplay", "SyncplayConsole.exe"
        )
else:
    syncplay_path = shutil.which("syncplay")

SYNCPLAY_PATH = syncplay_path

YTDLP_PATH = shutil.which("yt-dlp")  # already in pip deps

#########################################################################################

if __name__ == '__main__':
    pass



import re
import logging
import tempfile
from typing import Dict
import os
import json

import requests
from bs4 import BeautifulSoup

from aniworld.config import DEFAULT_REQUEST_TIMEOUT, MPV_SCRIPTS_DIRECTORY
from aniworld.common import copy_file_if_different, setup_autostart, setup_autoexit

CHAPTER_FORMAT = "\n[CHAPTER]\nTIMEBASE=1/1000\nSTART={}\nEND={}\nTITLE={}\n"
OPTION_FORMAT = "skip-{}_start={},skip-{}_end={}"
MAL_ANIME_URL = "https://myanimelist.net/anime/{}"
MAL_SEARCH_URL = "https://myanimelist.net/search/prefix.json?type=anime&keyword={}"
ANISKIP_API_URL = "https://api.aniskip.com/v1/skip-times/{}/{}?types=op&types=ed"


def ftoi(value: float) -> str:
    return str(int(value * 1000))


def check_episodes(anime_id):
    response = requests.get(
        MAL_ANIME_URL.format(anime_id),
        timeout=DEFAULT_REQUEST_TIMEOUT
    )
    soup = BeautifulSoup(response.content, 'html.parser')
    episodes_span = soup.find('span', class_='dark_text', string='Episodes:')

    if episodes_span and episodes_span.parent:
        episodes = episodes_span.parent.text.replace("Episodes:", "").strip()
        return int(episodes)

    logging.warning("The Number can not be found!")
    return None


def get_mal_id_from_title(title: str, season: int) -> int:
    logging.debug("Fetching MAL ID for: %s", title)

    name = re.sub(r' \(\d+ episodes\)', '', title)
    keyword = re.sub(r'\s+', '%20', name)

    response = requests.get(
        MAL_SEARCH_URL.format(keyword),
        timeout=DEFAULT_REQUEST_TIMEOUT
    )
    logging.debug("MyAnimeList response status code: %d", response.status_code)

    if response.status_code != 200:
        logging.error(
            "Failed to fetch MyAnimeList data. HTTP Status: %d", response.status_code
        )
        return 0

    mal_metadata = response.json()
    results = [
        entry for entry in mal_metadata['categories'][0]['items']
        if 'OVA' not in entry['name']
    ]

    if not results:
        logging.error("No match found!")
        raise ValueError("No match found!")

    best_match = results[0]
    anime_id = best_match['id']
    logging.debug("Found MAL ID: %s for %s", anime_id,
                  json.dumps(best_match, indent=4))

    while season > 1:
        anime_id = get_sequel_anime_id(anime_id)
        season -= 1

    return anime_id


def get_sequel_anime_id(anime_id: int) -> int:
    url = MAL_ANIME_URL.format(anime_id)
    response = requests.get(url, timeout=DEFAULT_REQUEST_TIMEOUT)
    response.raise_for_status()

    soup = BeautifulSoup(response.text, 'html.parser')
    sequel_div = soup.find(
        "div", string=lambda text: text and "Sequel" in text and "(TV)" in text)

    if not sequel_div:
        raise ValueError("Sequel (TV) not found")

    title_div = sequel_div.find_next("div", class_="title")
    if not title_div:
        raise ValueError("No 'title'-Div found")

    link_element = title_div.find("a")
    if not link_element:
        raise ValueError("No Link found in 'title'-Div")

    link_url = link_element.get("href")
    match = re.search(r'/anime/(\d+)', link_url)

    if not match:
        raise ValueError("No Anime-ID found in the link URL")

    return match.group(1)


def build_options(metadata: Dict, chapters_file: str) -> str:
    op_end, ed_start = None, None
    options = []

    for skip in metadata["results"]:
        skip_type = skip["skip_type"]
        st_time = skip["interval"]["start_time"]
        ed_time = skip["interval"]["end_time"]

        ch_name = "Opening" if skip_type == "op" else "Ending" if skip_type == "ed" else None
        if skip_type == "op":
            op_end = ed_time
        elif skip_type == "ed":
            ed_start = st_time

        with open(chapters_file, 'a', encoding='utf-8') as f:
            f.write(CHAPTER_FORMAT.format(
                ftoi(st_time), ftoi(ed_time), ch_name))

        options.append(OPTION_FORMAT.format(
            skip_type, st_time, skip_type, ed_time))

    if op_end:
        ep_ed = ed_start if ed_start else op_end
        with open(chapters_file, 'a', encoding='utf-8') as f:
            f.write(CHAPTER_FORMAT.format(
                ftoi(op_end), ftoi(ep_ed), "Episode"))

    return ",".join(options)


def build_flags(anime_id: str, episode: int, chapters_file: str) -> str:
    aniskip_api = ANISKIP_API_URL.format(anime_id, episode)
    response = requests.get(aniskip_api, timeout=DEFAULT_REQUEST_TIMEOUT)

    if response.status_code == 500:
        logging.info("Aniskip API is currently not working!")
        return ""
    if response.status_code != 200:
        logging.info("Failed to fetch AniSkip data.")
        return ""

    metadata = response.json()

    if not metadata.get("found"):
        logging.warning("No skip times found.")
        return ""

    with open(chapters_file, 'w', encoding='utf-8') as f:
        f.write(";FFMETADATA1")

    options = build_options(metadata, chapters_file)
    return f"--chapters-file={chapters_file} --script-opts={options}"


def aniskip(title: str, episode: int, season: int, aniworld_episodes: int) -> str:
    setup_autostart()
    setup_autoexit()
    setup_aniskip()

    anime_id = get_mal_id_from_title(
        title, season) if not title.isdigit() else title
    if not anime_id:
        logging.warning("No MAL ID found.")
        return ""

    if check_episodes(anime_id) == aniworld_episodes:
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as chapters_file:
            return build_flags(anime_id, episode, chapters_file.name)
    else:
        logging.warning("Mal ID isn't matching episode counter!")
        return ""


def setup_aniskip():
    script_directory = os.path.dirname(
        os.path.dirname(os.path.abspath(__file__)))
    mpv_scripts_directory = MPV_SCRIPTS_DIRECTORY

    if not os.path.exists(mpv_scripts_directory):
        os.makedirs(mpv_scripts_directory)

    skip_source_path = os.path.join(
        script_directory, 'aniskip', 'scripts', 'aniskip.lua')
    skip_destination_path = os.path.join(mpv_scripts_directory, 'aniskip.lua')

    copy_file_if_different(skip_source_path, skip_destination_path)


if __name__ == '__main__':
    # setup_aniskip()
    # setup_autoexit()
    # setup_autostart()

    print(get_mal_id_from_title("Kaguya-sama: Love is War", season=1))
